#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <inttypes.h>
#include <string.h>
#include <errno.h>

static void checked_fread(void *buffer, size_t size, FILE * infile)
{
    size_t rc = fread(buffer, 1, size, infile);

    if (rc != size)
    {
        int errnotemp = errno;
        long offset = ftell(infile);
        fflush(stdout);
        fprintf(stderr, "\nError reading 0x%lx bytes at offset 0x%lx: ",
                (unsigned long)size, (unsigned long)offset);
        if (feof(infile))
        {
            fprintf(stderr, "Unexpected end of file\n");
        }
        else
        {
            errno = errnotemp;
            perror(NULL);
        }
        fprintf(stderr, "\n");

        exit(-1);
    }
}

static void checked_fwrite(void *buffer, size_t size, FILE * outfile)
{
    size_t rc = fwrite(buffer, 1, size, outfile);

    if (rc != size)
    {
        int errnotemp = errno;
        fflush(stdout);
        fprintf(stderr, "\nError writing 0x%lx bytes: ", (unsigned long)size);
        errno = errnotemp;
        perror(NULL);
        fprintf(stderr, "\n");

        exit(-1);
    }
}

static void checked_fseek(FILE *infile, long offset, int whence)
{
    int rc = fseek(infile, offset, whence);

    if (rc != 0)
    {
        const char * whence_str = NULL;
        int errnotemp = errno;

        switch (whence)
        {
            case SEEK_SET:
                whence_str = "from beginning";
                break;
            case SEEK_CUR:
                whence_str = "from current position";
                break;
            case SEEK_END:
                whence_str = "from end";
                break;
            default:
                whence_str = "with INVALID \'WHENCE\'";
                break;
        }

        fflush(stdout);
        fprintf(stderr, "\nSeek to 0x%lx (%ld) %s failed: ",
                (unsigned long)offset, offset, whence_str);
        errno = errnotemp;
        perror(NULL);
        fprintf(stderr, "\n");

        exit(-1);
    }
}

FILE * checked_fopen(const char *file_name, const char *mode)
{
    FILE *file = fopen(file_name, mode);
    if (!file)
    {
        int errnotemp = errno;
        fprintf(stderr, "\nFailed opening %s: ", file_name);
        errno = errnotemp;
        perror(NULL);
        fprintf(stderr, "\n");

        exit(-1);
    }

    return file;
}

uint32_t read32le(FILE *infile)
{
    uint8_t buf[4];
    checked_fread(buf, 4, infile);
    
    return buf[0]*UINT32_C(0x1) |
           buf[1]*UINT32_C(0x100) |
           buf[2]*UINT32_C(0x10000) |
           buf[3]*UINT32_C(0x1000000);
}

int main(int argc, char **argv)
{
    FILE *infile, *outfile;
    uint32_t block_size;
    uint32_t bytes_per_channel;
    uint32_t channel_count;
    uint32_t start_offset;
    long file_size;

    if (argc != 3)
    {
        fprintf(stderr, "usage: unscrew infile.npsf outfiler.npsf\n");
        exit(-1);
    }

    infile = checked_fopen(argv[1], "rb");
    outfile = checked_fopen(argv[2], "wb");

    // get total file size for check
    checked_fseek(infile, 0, SEEK_END);
    file_size = ftell(infile);
    if (-1 == file_size)
    {
        perror("ftell");
    }
    checked_fseek(infile, 0, SEEK_SET);

    // read header

    {
        char head_buf[4];
        // 0x0: 'NPSF'
        checked_fread(head_buf, 4, infile);

        if (memcmp(head_buf, "NPSF", 4))
        {
            fprintf(stderr, "missing NPSF magic\n");
            exit(-1);
        }
    }

    // 0x4: block size
    block_size = read32le(infile);
    // 0x8: bytes per channel
    bytes_per_channel = read32le(infile);
    // 0xC: channels
    channel_count = read32le(infile);
    // 0x10: start offset
    start_offset = read32le(infile);

    // good enough, copy whole header (use start_offset as header size)
    {
        char *head_buf = malloc(start_offset);
        if (!head_buf)
        {
            perror("malloc failed");
            fprintf(stderr, "\n");
            exit(-1);
        }
        
        checked_fseek(infile, 0, SEEK_SET);
        checked_fread(head_buf, start_offset, infile);
        checked_fwrite(head_buf, start_offset, outfile);
    }

    {
        long total_bytes = (bytes_per_channel*channel_count + block_size - 1u) / block_size * block_size;
        long total_bytes_read = 0;
        long total_bytes_skipped = 0;

        for (long bytes_read = 0; bytes_read+16 <= total_bytes; bytes_read += 16)
        {
            char frame_buf[16];
            checked_fread(frame_buf, 16, infile);
            if (frame_buf[1] != 0)
            {
                int adjust;
                long curoffset = ftell(infile);

                for (adjust = 1; adjust <= 3; adjust ++)
                {
                    // Previous frame was too long, actual start of this frame
                    // is 16-adjust back, instead of the 16 back where we did the original
                    // read.

                    checked_fseek(infile, curoffset-(16-adjust), SEEK_SET);
                    checked_fread(frame_buf, 16, infile);
                    if (frame_buf[1] == 0)
                    {
                        break;
                    }
                }

                if (frame_buf[1] != 0)
                {
                    fprintf(stderr, "unable to find frame before 0x%lx\n", (unsigned long)curoffset);
                    exit(-1);
                }

                total_bytes_skipped += adjust;
                //printf("adjusted by %d at 0x%lx\n", adjust, (unsigned long)(curoffset-(16-adjust)));
            }

            total_bytes_read += 16;

            checked_fwrite(frame_buf, 16, outfile);
        }

        printf("\n\n");
        printf("   header bytes:  0x%"PRIx32" (expected 0x%lx)\n", start_offset, total_bytes);
        printf(" + body bytes:    0x%lx\n", total_bytes_read);
        printf(" + bytes skipped: 0x%lx\n", total_bytes_skipped);
        if (file_size == (start_offset + total_bytes_read + total_bytes_skipped))
        {
                printf("   ____________________\n");
        }
        else
        {
                printf(" !=____________________\n");
        }
        printf("   file size:     0x%lx\n", file_size);
    }
}
